@@ -11,12 +11,12 @@ use syn::Ident;
 
 /// The kind of tag to use.
 #[derive(Debug, Clone, Copy)]
-pub enum Tag {
+pub enum DefaultTag {
     Index,
     Name,
 }
 
-impl Default for Tag {
+impl Default for DefaultTag {
     fn default() -> Self {
         Self::Index
     }
@@ -50,10 +50,10 @@ impl Default for Packing {
 pub(crate) struct TypeAttr {
     /// `#[musli(tag_type)]`.
     pub(crate) tag_type: Option<(Span, syn::Type)>,
-    /// `#[musli(variant_tag)]`.
-    pub(crate) variant_tag: Tag,
-    /// `#[musli(field_tag)]`.
-    pub(crate) field_tag: Tag,
+    /// `#[musli(default_variant_tag = "..")]`.
+    pub(crate) default_variant_tag: DefaultTag,
+    /// `#[musli(default_field_tag = "..")]`.
+    pub(crate) default_field_tag: DefaultTag,
     /// `#[musli(packed)]` or `#[musli(transparent)]`.
     pub(crate) packing: Option<(Span, Packing)>,
 }
@@ -86,11 +86,11 @@ impl TypeAttr {
 #[derive(Default)]
 pub(crate) struct FieldAttr {
     /// Module to use when decoding.
-    pub(crate) encode_path: Option<(Span, syn::Path)>,
+    pub(crate) encode_path: Option<(Span, syn::ExprPath)>,
     /// Path to use when decoding.
-    pub(crate) decode_path: Option<(Span, syn::Path)>,
+    pub(crate) decode_path: Option<(Span, syn::ExprPath)>,
     /// Method to check if we want to skip encoding.
-    pub(crate) skip_encoding_if: Option<(Span, syn::Path)>,
+    pub(crate) skip_encoding_if: Option<(Span, syn::ExprPath)>,
     /// Rename a field to the given literal.
     pub(crate) rename: Option<(Span, syn::Expr)>,
     /// Use a default value for the field if it's not available.
@@ -101,10 +101,14 @@ pub(crate) struct FieldAttr {
 pub(crate) struct VariantAttr {
     /// `#[musli(tag_type)]`.
     pub(crate) tag_type: Option<(Span, syn::Type)>,
-    /// Rename a field to the given literal.
+    /// Rename a field to the given expression.
     pub(crate) rename: Option<(Span, syn::Expr)>,
     /// `#[musli(packed)]` or `#[musli(transparent)]`.
     pub(crate) packing: Option<(Span, Packing)>,
+    /// `#[musli(default)]`.
+    pub(crate) default: Option<Span>,
+    /// `#[musli(default_field_tag = "..")]`.
+    pub(crate) default_field_tag: Option<DefaultTag>,
 }
 
 impl VariantAttr {
@@ -157,12 +161,12 @@ impl FieldAttr {
     }
 
     /// Get skip encoding if.
-    pub(crate) fn skip_encoding_if(&self) -> Option<(Span, &syn::Path)> {
+    pub(crate) fn skip_encoding_if(&self) -> Option<(Span, &syn::ExprPath)> {
         let (span, path) = self.skip_encoding_if.as_ref()?;
         Some((*span, path))
     }
 
-    fn set_encode_path(&mut self, cx: &Ctxt, span: Span, encode_path: syn::Path) {
+    fn set_encode_path(&mut self, cx: &Ctxt, span: Span, encode_path: syn::ExprPath) {
         if self.encode_path.is_some() {
             cx.error_spanned_by(
                 encode_path,
@@ -173,7 +177,7 @@ impl FieldAttr {
         }
     }
 
-    fn set_decode_path(&mut self, cx: &Ctxt, span: Span, decode_path: syn::Path) {
+    fn set_decode_path(&mut self, cx: &Ctxt, span: Span, decode_path: syn::ExprPath) {
         if self.decode_path.is_some() {
             cx.error_spanned_by(
                 decode_path,
@@ -184,7 +188,7 @@ impl FieldAttr {
         }
     }
 
-    fn set_skip_encoding_if(&mut self, cx: &Ctxt, span: Span, skip_encoding_if: syn::Path) {
+    fn set_skip_encoding_if(&mut self, cx: &Ctxt, span: Span, skip_encoding_if: syn::ExprPath) {
         if self.skip_encoding_if.is_some() {
             cx.error_spanned_by(
                 skip_encoding_if,
@@ -212,12 +216,12 @@ pub(crate) fn type_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> TypeAttr {
                         attr.tag_type = Some((path.span(), ty));
                     }
                 }
-                // parse #[musli(variant_tag = "..")]
-                Attribute::KeyValue(path, expr) if path == VARIANT => {
-                    if let Some(tag) = parse_value_string(cx, VARIANT, expr) {
-                        attr.variant_tag = match tag.value().as_str() {
-                            "index" => Tag::Index,
-                            "name" => Tag::Name,
+                // parse #[musli(default_variant_tag = "..")]
+                Attribute::KeyValue(path, expr) if path == DEFAULT_VARIANT_TAG => {
+                    if let Some(tag) = parse_value_string(cx, DEFAULT_VARIANT_TAG, expr) {
+                        attr.default_variant_tag = match tag.value().as_str() {
+                            "index" => DefaultTag::Index,
+                            "name" => DefaultTag::Name,
                             _ => {
                                 cx.error_spanned_by(
                                     tag,
@@ -228,12 +232,12 @@ pub(crate) fn type_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> TypeAttr {
                         };
                     }
                 }
-                // parse #[musli(field_tag = "..")]
-                Attribute::KeyValue(path, expr) if path == FIELD => {
-                    if let Some(tag) = parse_value_string(cx, FIELD, expr) {
-                        attr.field_tag = match tag.value().as_str() {
-                            "index" => Tag::Index,
-                            "name" => Tag::Name,
+                // parse #[musli(default_field_tag = "..")]
+                Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_TAG => {
+                    if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_TAG, expr) {
+                        attr.default_field_tag = match tag.value().as_str() {
+                            "index" => DefaultTag::Index,
+                            "name" => DefaultTag::Name,
                             _ => {
                                 cx.error_spanned_by(
                                     tag,
@@ -279,6 +283,7 @@ pub(crate) fn field_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> FieldAttr {
                         let mut encode_path = path.clone();
 
                         encode_path
+                            .path
                             .segments
                             .push(Ident::new("encode", Span::call_site()).into());
 
@@ -287,6 +292,7 @@ pub(crate) fn field_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> FieldAttr {
                         let mut decode_path = path.clone();
 
                         decode_path
+                            .path
                             .segments
                             .push(Ident::new("decode", Span::call_site()).into());
 
@@ -361,6 +367,22 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
                         }
                     }
                 }
+                // parse #[musli(default_field_tag = "..")]
+                Attribute::KeyValue(path, expr) if path == DEFAULT_FIELD_TAG => {
+                    if let Some(tag) = parse_value_string(cx, DEFAULT_FIELD_TAG, expr) {
+                        attr.default_field_tag = Some(match tag.value().as_str() {
+                            "index" => DefaultTag::Index,
+                            "name" => DefaultTag::Name,
+                            _ => {
+                                cx.error_spanned_by(
+                                    tag,
+                                    format!("illegal #[{}({})] value", ATTR, TAG),
+                                );
+                                continue;
+                            }
+                        });
+                    }
+                }
                 // parse #[musli(transparent)]
                 Attribute::Path(path) if path == TRANSPARENT => {
                     attr.set_packing(cx, path.span(), Packing::Transparent);
@@ -369,6 +391,10 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
                 Attribute::Path(path) if path == PACKED => {
                     attr.set_packing(cx, path.span(), Packing::Packed);
                 }
+                // parse #[musli(default)]
+                Attribute::Path(path) if path == DEFAULT => {
+                    attr.default = Some(path.span());
+                }
                 meta => {
                     cx.error_span(meta.span(), format!("unsupported #[{}] attribute", ATTR));
                 }
@@ -380,7 +406,7 @@ pub(crate) fn variant_attrs(cx: &Ctxt, attrs: &[syn::Attribute]) -> VariantAttr
 }
 
 /// Get expression path.
-fn value_as_path<'a>(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<syn::Path> {
+fn value_as_path<'a>(cx: &Ctxt, attr: Symbol, value: AttributeValue) -> Option<syn::ExprPath> {
     match value {
         AttributeValue::Path(path) => Some(path),
         _ => {
@@ -452,7 +478,7 @@ where
 /// The flexible value of an attribute.
 pub enum AttributeValue {
     /// A path.
-    Path(syn::Path),
+    Path(syn::ExprPath),
     /// A type.
     Type(syn::Type),
     /// A literal value.
@@ -505,10 +531,10 @@ impl Parse for TypeAttributes {
                 let value = match &path {
                     // parse #[musli(tag_type = <type>)]
                     path if path == TAG_TYPE => AttributeValue::Type(content.parse()?),
-                    // parse #[musli(variant = "..")]
-                    path if path == VARIANT => AttributeValue::Lit(content.parse()?),
-                    // parse #[musli(field_tag = "..")]
-                    path if path == FIELD => AttributeValue::Lit(content.parse()?),
+                    // parse #[musli(default_variant_tag = "..")]
+                    path if path == DEFAULT_VARIANT_TAG => AttributeValue::Lit(content.parse()?),
+                    // parse #[musli(default_field_tag = "..")]
+                    path if path == DEFAULT_FIELD_TAG => AttributeValue::Lit(content.parse()?),
                     path => {
                         return Err(syn::Error::new(path.span(), "unsupported attribute"));
                     }
@@ -552,6 +578,8 @@ impl Parse for VariantAttributes {
                     path if path == TAG_TYPE => AttributeValue::Type(content.parse()?),
                     // parse #[musli(tag = <expr>)]
                     path if path == TAG => AttributeValue::Expr(content.parse()?),
+                    // parse #[musli(default_field_tag = <expr>)]
+                    path if path == DEFAULT_FIELD_TAG => AttributeValue::Lit(content.parse()?),
                     path => {
                         return Err(syn::Error::new(path.span(), "unsupported attribute"));
                     }
